/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef elf64_h
#define elf64_h

#ifndef ELF_CLASS
#define ELF_CLASS	ELFCLASS64
#endif
#include "elfxx.h"

#endif /* elf64_h */
