/* -*- Mode: IDL; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

/* DOM stubs so that we can pass DOM objects across XPConnect */

%{C++
/* C++ should ignore this file because there are real DOM interfaces
   elsewhere, so if 0 it out
*/
#if 0
%}

#include "domstubs.idl"

[scriptable, uuid(7FB67F1D-F1FF-450A-9A63-261344E13AC4)]
interface nsIDOMXULElement : nsIDOMElement {};

[scriptable, uuid(E3F0D263-10A2-4703-BD5F-E68F32F4DC31)]
interface nsIDOMXULTreeElement : nsIDOMXULElement {};

[scriptable, uuid(f3c50361-14fe-11d3-bf87-00105a1b0627)]
interface nsIDOMXULCommandDispatcher : nsISupports {};

[scriptable, uuid(17ddd8c0-c5f8-11d2-a6ae-00104bde6048)]
interface nsIDOMXULDocument : nsISupports {};

%{C++
#endif
%}
