/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef nsCP1125ToUnicode_h___
#define nsCP1125ToUnicode_h___

#include "nsISupports.h"
/**
 * A character set converter from CP1125 to Unicode.
 */
nsresult
nsCP1125ToUnicodeConstructor(nsISupports *aOuter, REFNSIID aIID,
                            void **aResult);


#endif /* !nsCP1125ToUnicode_h___ */
